from io import StringIO, BytesIO

from safe_radix32.pure import SAFE_MAP, SAFE_BASE_BITS

DECODE_UNPAD = {1: 4, 3: 3, 4: 2, 6: 1}
DECODE_REVERSE_MAP = {v: k for k, v in enumerate(SAFE_MAP.encode())}
ENCODE_DOUBLED_MAP = [a + b for a in SAFE_MAP for b in SAFE_MAP]


def encode_safe_radix32_bytes(b: bytes) -> str:
    leftover = len(b) % 5
    buffer = StringIO()
    _fb = int.from_bytes
    for i in range(0, len(b) - leftover, 5):
        v = _fb(b[i : i + 5], "big")
        buffer.write(
            "".join(
                (
                    ENCODE_DOUBLED_MAP[v >> 30],
                    ENCODE_DOUBLED_MAP[(v >> 20) & 1023],
                    ENCODE_DOUBLED_MAP[(v >> 10) & 1023],
                    ENCODE_DOUBLED_MAP[v & 1023],
                )
            )
        )
    if leftover:
        pad = 5 - leftover
        v = _fb(b[-leftover:] + b"\x00" * pad, "big")
        buffer.write(ENCODE_DOUBLED_MAP[v >> 30])  # 10 : 1
        if pad < 4:
            buffer.write(ENCODE_DOUBLED_MAP[(v >> 20) & 1023])  # 20 : 2
        if pad < 3:
            buffer.write(SAFE_MAP[(v >> 15) & 31])  # 25 : 3
        if pad < 2:
            buffer.write(ENCODE_DOUBLED_MAP[(v >> 5) & 1023])  # 35 : 4
        # buffer.append(SAFE_MAP[v & 31])
    return buffer.getvalue()


def decode_safe_radix32_bytes(s: str) -> bytes:
    s = s.encode("ascii")
    leftover = len(s) % 8
    buffer = BytesIO()
    try:
        for o in range(0, len(s) - leftover, 8):
            u = 0
            for c in s[o : o + 8]:
                u = (u << SAFE_BASE_BITS) + DECODE_REVERSE_MAP[c]
            buffer.write(u.to_bytes(5, "big"))
        if leftover:
            pad = 8 - leftover
            s = s[-leftover:] + b"2" * pad
            u = 0
            for c in s:
                u = (u << SAFE_BASE_BITS) + DECODE_REVERSE_MAP[c]
            b = u.to_bytes(5, "big")
            buffer.write(b[: DECODE_UNPAD[pad]])
    except KeyError:
        raise OverflowError
    return buffer.getvalue()



"""
safe_radix32
Radix32 encode long integers with a safe alphabet.

:copyright: 2026 Nándor Mátravölgyi
:license: Apache2, see LICENSE for more details.
"""
import os

__author__ = "Nándor Mátravölgyi"
__copyright__ = "Copyright 2026 Nándor Mátravölgyi"
__author_email__ = "nandor.matra@gmail.com"
__version__ = "0.4.1"


if os.environ.get("SAFE_RADIX32_PUREPYTHON"):
    from .pure import (
        encode_safe_radix32 as encode,
        decode_safe_radix32 as decode,
        SAFE_RADIX32_ALPHABET as ALPHABET,
        SAFE_RADIX32_ALPHABET_RE as ALPHABET_RE,
        encode_safe_radix32_fixed_width as encode_fw,
    )
else:
    try:
        from ._cython import (
            encode_safe_radix32 as encode,
            decode_safe_radix32 as decode,
            SAFE_RADIX32_ALPHABET as ALPHABET,
            SAFE_RADIX32_ALPHABET_RE as ALPHABET_RE,
            encode_safe_radix32_fixed_width as encode_fw,
        )
    except ImportError:
        from .pure import (
            encode_safe_radix32 as encode,
            decode_safe_radix32 as decode,
            SAFE_RADIX32_ALPHABET as ALPHABET,
            SAFE_RADIX32_ALPHABET_RE as ALPHABET_RE,
            encode_safe_radix32_fixed_width as encode_fw,
        )



import os

from safe_radix32.bytes import encode_safe_radix32_bytes, decode_safe_radix32_bytes


def test_safe_radix32_bytes():
    # test translation integrity with some patterns
    for length in range(1, 40):
        for start in range(256):
            for step in [1, 2, 34, 8, 13, 21, 31, 32, 55]:
                v = b"".join([((i * step + start) % 256).to_bytes(1, "big") for i in range(length)])
                e = encode_safe_radix32_bytes(v)
                assert decode_safe_radix32_bytes(e) == v, f"{v}\n{e}"
        for _ in range(1000):
            v = os.urandom(length)
            e = encode_safe_radix32_bytes(v)
            assert decode_safe_radix32_bytes(e) == v, f"{v}\n{e}"

    rb = b"\x01\x00+\xdcTb\x91\xf4\xb1'\x10\xd5~\x15\xce\n\x05V\x07@M_\xb4\xf9\x8b"
    e = encode_safe_radix32_bytes(rb)
    assert e == "2724gg4bKGCzFKFB46Ggq8MM3C4cK3m2FcPpFwKJ"
    assert decode_safe_radix32_bytes(e) == rb



import random

import pytest

import safe_radix32


def _to_fixed_width(t: str) -> str:
    return safe_radix32.ALPHABET[0] * (13 - len(t)) + t


def test_safe_radix32_base():
    assert len(safe_radix32.encode_fw(0)) == 13
    assert safe_radix32.ALPHABET == "2346789BCFGJKLMPQVWZbcfgjkmpqvwz"
    for i in range(32):
        t = safe_radix32.encode(i)
        assert t == safe_radix32.ALPHABET[i]
        assert safe_radix32.ALPHABET_RE.match(t) is not None
        assert safe_radix32.decode(t) == i
    for i, t in [
        (2 ** 32 - 1, "6zzzzzz"),
        (2 ** 32, "7222222"),
        (2 ** 48 - 1, "Bzzzzzzzzz"),
        (2 ** 48, "C222222222"),
        (2 ** 60 - 1, "zzzzzzzzzzzz"),
        (2 ** 61 - 1, "3zzzzzzzzzzzz"),
        (2 ** 62 - 1, "6zzzzzzzzzzzz"),
        (2 ** 63 - 1, "Bzzzzzzzzzzzz"),
        (-(2 ** 32) + 1, "Pzzzzzq222223"),
        (-(2 ** 32), "Pzzzzzq222222"),
        (-(2 ** 62), "K222222222222"),
        (-(2 ** 63), "C222222222222"),
        (-(2 ** 63) + 1, "C222222222223"),
        (int("01" * 32, 2), "8GcGcGcGcGcGc"),
        (int("10" * 31, 2), "4cGcGcGcGcGcG"),
        (int("0110" * 16, 2), "9KkZ9KkZ9KkZ9"),
        (int("0011" * 16, 2), "69KkZ9KkZ9KkZ"),
        (int("100" + "1100" * 15, 2), "7kZ9KkZ9KkZ9K"),
        (8474785761110355227, "BJB6qgwJCGpCp"),
        (3927410648750184479, "6L2Bg8bwvMf2z"),
    ]:
        t_fw = _to_fixed_width(t)
        assert safe_radix32.encode(i) == t, t
        assert safe_radix32.encode_fw(i) == t_fw, f"{i} {t_fw}"
        assert safe_radix32.decode(t) == i, t
        assert safe_radix32.ALPHABET_RE.match(t) is not None
    # random consistency checks
    for _ in range(5000):
        r = random.randint(-(2 ** 63), 2 ** 63 - 1)
        x = safe_radix32.encode(r)
        x_fw = _to_fixed_width(x)
        assert safe_radix32.encode_fw(r) == x_fw
        assert safe_radix32.decode(x) == r, x
        assert safe_radix32.ALPHABET_RE.match(x) is not None


def test_safe_radix32_limits():
    with pytest.raises(OverflowError):
        safe_radix32.encode(2 ** 63)
    with pytest.raises(OverflowError):
        safe_radix32.encode_fw(2 ** 63)
    with pytest.raises(OverflowError):
        safe_radix32.encode(-(2 ** 63) - 1)
    with pytest.raises(OverflowError):
        safe_radix32.encode_fw(-(2 ** 63) - 1)
    for c in "01AaEIoOyYxXrR/*_:?#&%!+'\"":
        with pytest.raises(OverflowError):
            safe_radix32.decode(c)
    for c in "\xffáŐ":
        with pytest.raises(UnicodeEncodeError):
            safe_radix32.decode(c)
    assert safe_radix32.decode("222") == 0
    assert safe_radix32.decode("2222222222222") == 0
    assert safe_radix32.decode("7777") == 135300
    assert safe_radix32.decode("7777777777777") == 4760450083537948804
    assert safe_radix32.decode("3777777777777") == 1301685569717407876
    assert safe_radix32.decode("zzzzzzzzzzzz") == 1152921504606846975
    assert safe_radix32.decode("Pzzzzzzzzzzzz") == -1
    for v in ["88888888888888", "48888888888888", "Qzzzzzzzzzzzz", "zzzzzzzzzzzzz"]:
        with pytest.raises(OverflowError):
            safe_radix32.decode(v)


def test_safe_radix32_types():
    for v in [None, ["88"], 5451]:
        with pytest.raises(TypeError):
            safe_radix32.decode(v)
    for v in [None, ["88"], "8888"]:
        with pytest.raises(TypeError):
            safe_radix32.encode(v)
        with pytest.raises(TypeError):
            safe_radix32.encode_fw(v)



import os
import sys

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

PYPY = hasattr(sys, "pypy_version_info")


class BuildExt(build_ext):
    def build_extension(self, ext):
        for source in ext.sources:
            pyx = os.path.splitext(source)[0] + ".pyx"
            if not os.path.exists(pyx):
                if not cythonize:
                    print("WARNING")
                    print("Cython is required for building extension.")
                    print("Cython can be installed from PyPI.")
                    print("Falling back to pure Python implementation.")
                    return
                cythonize(pyx)
            elif os.path.exists(pyx) and os.stat(source).st_mtime < os.stat(pyx).st_mtime and cythonize:
                cythonize(pyx)
        try:
            return build_ext.build_extension(self, ext)
        except Exception as e:
            print("WARNING: Failed to compile extension modules.")
            print("Falling back to pure Python implementation.")
            print(e)


class Sdist(sdist):
    def __init__(self, *args, **kwargs):
        cythonize("safe_radix32/_cython.pyx")
        sdist.__init__(self, *args, **kwargs)


ext_modules = []
if not PYPY:
    ext_modules.append(Extension("safe_radix32._cython", sources=["safe_radix32/_cython.c"]))

setup(
    packages=["safe_radix32"],
    cmdclass={"build_ext": BuildExt, "sdist": Sdist},
    ext_modules=ext_modules,
)
